@@ -270,7 +270,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
             self.focus_in_events.append(handler_id)
             handler_id = treeview.connect("focus-out-event", self.on_treeview_focus_out_event)
             self.focus_out_events.append(handler_id)
-        self.prev_path, self.next_path = None, None
+        self.current_path, self.prev_path, self.next_path = None, None, None
         self.on_treeview_focus_out_event(None, None)
         self.treeview_focussed = None
 
@@ -437,6 +437,7 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
     def set_locations(self, locations):
         self.set_num_panes(len(locations))
         locations = [os.path.abspath(l or ".") for l in locations]
+        self.current_path = None
         self.model.clear()
         for pane, loc in enumerate(locations):
             self.fileentry[pane].set_filename(loc)
@@ -714,13 +715,22 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         if not cursor_path:
             self.emit("next-diff-changed", False, False)
         else:
-            # TODO: Only recalculate when needed;
-            # not self.prev_path < cursor_path < self.next_path is a start,
-            # but fails when we move off a changed row.
-            prev_path, next_path = self.model._find_next_prev_diff(cursor_path)
-            self.prev_path, self.next_path = prev_path, next_path
-            have_next_diffs = (prev_path is not None, next_path is not None)
-            self.emit("next-diff-changed", *have_next_diffs)
+            if self.current_path:
+                old_cursor = self.model.get_iter(self.current_path)
+                state = self.model.get_state(old_cursor, 0)
+                # We can skip recalculation if the new cursor is between the
+                # previous/next bounds, and we weren't on a changed row
+                skip = state in (tree.STATE_NORMAL, tree.STATE_EMPTY) and \
+                       self.prev_path < cursor_path < self.next_path
+            else:
+                skip = False
+
+            if not skip:
+                prev, next = self.model._find_next_prev_diff(cursor_path)
+                self.prev_path, self.next_path = prev, next
+                have_next_diffs = (prev is not None, next is not None)
+                self.emit("next-diff-changed", *have_next_diffs)
+        self.current_path = cursor_path
 
         paths = self._get_selected_paths(pane)
         if len(paths) > 0:
@@ -975,22 +985,24 @@ class DirDiff(melddoc.MeldDoc, gnomeglade.Component):
         return True
 
     def on_treeview_button_press_event(self, treeview, event):
-        # unselect other panes
-        for t in filter(lambda x:x!=treeview, self.treeview[:self.num_panes]):
+        # Unselect any selected files in other panes
+        for t in [v for v in self.treeview[:self.num_panes] if v != treeview]:
             t.get_selection().unselect_all()
+
         if event.button == 3:
-            try:
-                path, col, cellx, celly = treeview.get_path_at_pos( int(event.x), int(event.y) )
-            except TypeError:
-                pass # clicked outside tree
-            else:
-                treeview.grab_focus()
-                selected = self._get_selected_paths( self.treeview.index(treeview) )
-                if len(selected) <= 1 and event.state == 0:
-                    treeview.set_cursor( path, col, 0)
-                self.popup_in_pane(self.treeview.index(treeview), event)
-            return event.state==0
-        return 0
+            treeview.grab_focus()
+            path = treeview.get_path_at_pos(int(event.x), int(event.y))
+            selection = treeview.get_selection()
+            model, rows = selection.get_selected_rows()
+
+            if path[0] not in rows:
+                selection.unselect_all()
+                selection.select_path(path[0])
+                treeview.set_cursor(path[0])
+
+            self.popup_in_pane(self.treeview.index(treeview), event)
+            return True
+        return False
 
     def get_state_traversal(self, diffmapindex):
         def tree_state_iter():
